/*
    This file implements and tests speed of
    a faster modulo for fixed m, maybe.
    
    Using GMP to benchmark against the fast REDC and existing implementations
    Positively faster above REDC boundaries it seems when FFT kicks in...
    
    Author: Vincent Diepeveen
    email: diep at xs4all dot nl
    
    note: programmed at my macbookpro with os/x
          seems including header files of gmp is
          a problem there. I compile it with:
          
          /gmp/> cat makedm
          #!/bin/sh
          gcc -O2 -o dm diepmod.c /gmp/gmp-4.2.2/.libs/libgmp.a
          /gmp/> 

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "gmp-4.2.2/gmp.h" // perhaps remove directory name at linux
                           // as i'm os/x at home 

#define UNIX 1  // windoze boys have to fill in here what they are

gmp_randstate_t randomstate;

int GetClock(void) {
  #if UNIX
  //return(clock());
  struct timeval timeval;
  gettimeofday(&timeval,NULL);
  return((int)(timeval.tv_sec*1000+(timeval.tv_usec/1000)));
  #else
  return((int)GetTickCount());
  #endif
}

int main(void) 
{
  mpz_t p,q,n,M,t,res1,res2,res2a,res2b,res2c,s,pow2i,squaredM;
  int sizen,sizep,sizeq,sizeM,endcompare,t1,t2,i,testlen,coll;
  gmp_randinit_mt(randomstate);
  
   printf("    size of 1 limb              =  %i bytes\n", 
   (int)sizeof(mp_limb_t));
/*
   in RSA we have      n = p * q
   where p,q are both large 'industry grade' primes,
   therefore we have an odd n
   
   and we have message M that we want to encode with a big E,
   
   C = M ^ E  (mod n)
   
   Now we repeat this for a lot of texts M of course.
   
   So the interesting thing is to take a faster modulo n,  
   where n is a constant throughout the entire encryption
   of the entire data that we want to encode.
     
   Focus of this file is just generating a method to take that modulo faster
   than O ( n ^ 2 ) which is what REDC and division do currently in GMP.
   
   Let's first generate therefore p,q
     n = p.q
   
   
   
   The modulo idea explained:
   
*/   
   mpz_init(p);
   mpz_init(q);
   mpz_init(n);
   mpz_init(M);
   mpz_init(res1);
   mpz_init(res2);
   mpz_init(res2a);
   mpz_init(res2b);
   mpz_init(res2c);
   mpz_init(s);
   mpz_init(pow2i);
   mpz_init(t);
   mpz_init(squaredM);
   
   // results we have we collect as a summation in res1 and res2
   mpz_init(res1);
   mpz_init(res2); 
  
   testlen =  1024*1024;
   sizen = 128; // let's start with 128 bits to see whether it works bugfree
               
  for( sizen = 128; sizen <= 128*16384; sizen *= 2 ) {              
   testlen >>= 1; 
   sizep = (sizen * 55) / 100; // set sizep at 55% size of n
   sizeq = sizen - sizep;      // sizeq is 45% of n
   sizeM = sizen - 2;          // our message is 2 bits smaller than n 
  
  // let's generate randomly p,q and then set it to first prime
   mpz_urandomb(p,randomstate,sizep);
   //mpz_nextprime(p,p);
   mpz_setbit(p,0);
   mpz_urandomb(q,randomstate,sizeq);
   //mpz_nextprime(q,q);
   mpz_setbit(q,0); 
   
   mpz_mul(n,p,q); 
   // n = p.q    
   // and just like in RSA our n is hard to factorize
   // well, that is, for most of us
   
   // precalculate s = 2^2i / n
   //  where 'i' is the number of bits of n
   mpz_set_ui(pow2i,1);
   mpz_mul_2exp(pow2i,pow2i,sizen+sizen);
   mpz_fdiv_q(s,pow2i,n);

   mpz_urandomb(M,randomstate,sizeM);
    // now make sure M is odd
   mpz_setbit(M,0);      
   
   // square M
   mpz_mul(M,M,M);
   mpz_set(squaredM,M);
   
   // the correct GMP calculation
   // M = res1 (mod n) 
   t1 = GetClock();
   for( i = 0; i < testlen; i++ ) 
   {
     mpz_add_ui(squaredM,squaredM,2);
     mpz_mod(res1,squaredM,n);
   }
   t2 = GetClock();
   coll = t2-t1;
   printf("bits = %i time original code = %.03lf seconds\n",
    sizen,((double)coll)/1000.00); 
   fflush(stdout);

   // get me the quotient also
   //mpz_fdiv_q(t,M,n);

   // the Diep approximation calculation
   // M = res2 (mod n)      ;
   // res2 >>= sizen        ; rightshift res2

   mpz_set(squaredM,M);
   t1 = GetClock();
   for( i = 0; i < testlen; i++ ) 
   {
      mpz_add_ui(squaredM,squaredM,2);
      mpz_mul(res2a,squaredM,s);
      mpz_fdiv_q_2exp(res2b,res2a,sizen+sizen);
     //printf(" quotient benadering = ");
     //mpz_out_str(stdout,10,res2);
     //printf("\n");
     mpz_mul(res2c,res2b,n);
     mpz_sub(res2,squaredM,res2c);
     for( ;; ) {
       int intres = mpz_cmp(res2,n);
       if( intres >= 0 )
       {
         //printf("needed 1 correction\n");
         mpz_sub(res2,res2,n);
       } 
       else
       {
         break;
       } 
     }
   }    
   t2 = GetClock();
   coll = t2-t1;
   printf("bits = %i time Diep Reduction = %.03lf seconds\n",
    sizen,((double)coll)/1000.00);
   fflush(stdout); 
  // above is our mpz_diepmod(res2,M,n,s);
   
   endcompare = mpz_cmp(res1,res2);
   if( endcompare == 0 ) 
     printf("both modulo's were equal and correct therefore\n");
   else
     printf("modulo has an incorrect outcome\n");
   fflush(stdout);
  }   
   return 0;
}
